<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
  "http://www.w3.org/TR/html4/loose.dtd">
<html{% if language is not none %} lang="{{ language }}"{% endif %}>
  <head>
    <title>{% trans filename=filename, docstitle=docstitle|e %}{{ filename }} &#8212; {{ docstitle }}{% endtrans %}</title>
    <style type="text/css">
      .hl { background-color: yellow }
    </style>
  </head>
  <body style="font-size: 90%">
    <pre>
      {{ text }}
    </pre>
  </body>
</html>
